/*
** Module   : Label.java
** Abstract : Label widget implementation.
**
** Copyright (c) 2005-2015, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ----------------------------Description------------------------------
** 001 SVG 20050609   @21452 Initial version.
** 003 SIY 20050831   @22506 Refactoring (see @@22394-22409).
** 004 SIY 20050907   @22640 Fixed use of @inheritDoc tag in javadoc.
** 005 SIY 20050926   @22842 Using ColorTable for drawing.
** 006 SIY 20060209   @24471 Updated to accommodate changes in xxxConfig
**                           classes. Minor cleanups.
** 007 SIY 20060323   @25222 Added copy constructor.
** 008 SIY 20060615   @27239 Do not draw label if its parent is null.
** 009 SIY 20060824   @28853 Removed getP2JType() method.
** 010 SIY 20060908   @29380 Fixed drawing.
** 011 GES 20061013   @30401 Added a delimiter flag and changed drawing in redirected output mode
**                           (if we are marked as a delimiter) to avoid the output of the line 
**                           itself (instead we only output a blank line).
** 012 SIY 20070214   @32128 Renamed getDelimiter() to isDelimiter() to match common naming 
**                           convention for boolean properties.
** 013 SIY 20071004   @35366 Rolled back #011.
** 014 NVS 20080401   @37778 Added isFocusable() method implementation.
** 015 SIY 20080401   @38798 Added support for quirks label positioning in ATTR-SPACE handling.
** 016 SIY 20100111   @44522 Fixed a rare case when drawing of empty space instead of label is 
**                           necessary. 
** 017 SIY 20100619          Switching to own implementation of character UI.
** 018 SIY 20100804          Changes in imports, some refactorings.
** 019 SIY 20100822          Refactored AttributeContainer support.
** 020 SIY 20100915          Renamed base class.
** 021 SIY 20110112          Changes related to new color handling scheme.
** 022 SIY 20111101          Updated imports.
** 023 SIY 20111202          Refactored and moved from CHUI package.
** 024 CA  20140805          Added generics to force the abstract widget classes to not use
**                           implementation-specific code: the abstract widgets will have access
**                           only to the base APIs provided by the OutputManager.
** 025 HC  20140911          GUI coordinate support - data types refactoring.
** 026 CA  20140926          Added shared widget configuration support, to allow GUI/ChUI
**                           concrete implementation for the same widget ID.  Refactored the
**                           widget configuration classes: all fields were made public; these
**                           classes need to be as dumb as possible, all logic related to
**                           setting/getting a certain field should be at the widget or at the
**                           caller. Refs #2254
** 027 HC  20141021          Implemented window move/position support.
** 028 CA  20141220          Implemented drawCaret, as is a no-op for LABEL widget.
** 029 CA  20150105          Changes for SIDE-LABEL-HANDLE runtime support.
** 030 CA  20150114          Fixed support for changing the ROW/COLUMN attrs from server-side, 
**                           for a side-label.
** 031 CA  20150127          If label text is changed, invalidate the location.
** 032 HC  20150128          Added support for frame widget position changes induced by
**                           setting ROW/COL/X/Y attributes.
*/

package com.goldencode.p2j.ui.client;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.client.widget.*;

/**
 * Implementation of label for Progress components. Progress uses TEXT
 * components for labels but this seems to be incorrect. TEXT implies
 * underlying variable and complex format.
 */
public abstract class Label<O extends OutputManager<?>>
extends FixedSizeContainer<O>
{
   /** Widget's size which is set by <code>setSize</code> method. */
   private boolean forcedSize = false;
   
   /** Component attribute container. */
   private BaseConfig config = null;

   /** Make label aligned to right boundary. */
   protected boolean rightAligned = true;

   /** Display additional line */
   protected boolean underlined = false;
   
   /** Flag the special case of use as a separator line. */
   private boolean delimiter = false;

   /** Label right shift offset. */
   protected double offset = 0;

   /** If it is <code>true</code>, empty string is displayed.*/
   protected boolean emptyMode = false;
   
   /** Prepare for disabling empty mode. */
   protected boolean disableEmptyMode = false;
   
   /**
    * Constructs label with given text.
    * 
    * @param   id
    *          The widget's ID.
    * @param   text
    *          Initial text.
    */
   public Label(WidgetId id, String text)
   {
      config = ConfigManager.getInstance().resolveWidgetConfig(id, LabelConfig.class);

      setName(text);
   }

   /**
    * Special constructor used to restore component using given configuration.
    * <p>
    * Call {@link #initialize} to finish the widget initialization.
    * 
    * @param   id
    *          The widget's ID.
    * @param   cfg
    *          Configuration data.
    */
   public Label(WidgetId id, WidgetConfig cfg)
   {
   }
   
   /**
    * Replacement API for the logic which should exist in the widget's c'tor, but the logic 
    * can't be executed as it depends on the widget being registered into the official registry.
    * <p>
    * This API is called after the widget is created and added to the registry.
    * 
    * @param    id
    *           The widget's ID.
    * @param    cfg
    *           The config used to initialize this widget.
    */
   @Override
   public void initialize(WidgetId id, WidgetConfig cfg)
   {
      config = (BaseConfig) ConfigManager.getInstance().resolveWidgetConfig(id, cfg);
      ConfigManager.getInstance().replaceConfig(config, cfg);
   }

   /**
    * Create a copy of the label.
    * 
    * @param   id
    *          The widget's ID.
    * @param   label
    *          Label to copy parameters from.
    */
   public Label(WidgetId id, Label<O> label)
   {
      config = ConfigManager.getInstance().resolveWidgetConfig(id, LabelConfig.class);
      setName(label.config.name);
      
      setUnderlined(label.isUnderlined());
      setRightAligned(label.isRightAligned());
      setWidth(label.width());
      forcedSize = label.forcedSize;
      setLocation(label.location());
      setVisible(label.isVisible());
   }

   /**
    * Provide access to the attributes stored in the widget.
    * 
    * @return  Reference to attribute container.
    */
   @SuppressWarnings("unchecked")
   @Override
   public BaseConfig config()
   {
      return config;
   }

   /**
    * Marks this label as one used to draw a separator line for header
    * widgets.  In this case there is no label text but the width is
    * positive and the underline attribute is enabled.
    *
    * @param    delimiter
    *           <code>true</code> to set this as a separator line.
    */
   public void setDelimiter(boolean delimiter)
   {
      this.delimiter = delimiter;
   }
   
   /**
    * Gets the flag that marks this label as one used to draw a separator
    * line for header widgets.
    *
    * @return   <code>true</code> if this label is a separator line.
    */
   public boolean isDelimiter()
   {
      return delimiter;
   }

   /**
    * Draws text. Screen value is left justified or padded to the right bound
    * with spaces. Non first lines are filled with spaces.
    */
   @Override
   public abstract void draw();

   /**
    * Sets position of terminal's cursor. Really does nothing because this
    * widget will never get an input focus.
    */
   @Override
   public void drawCaret()
   {
   }

   /**
    * Sets new widget's size. After call to this method the widget will
    * not change size automatically accordingly to screen value changes.
    * 
    * @param   width
    *          New widget width.
    * @param   height
    *          New widget height.
    */
   @Override
   public void setSize(double width, double height)
   {
      forcedSize = true;
      super.setSize(width, height);
   }

   /**
    * Set widget width.
    * 
    * @param   width
    *          New widget width.
    */
   @Override
   public void setWidth(double width)
   {
      forcedSize = true;
      super.setWidth(width);
   }
   
   /**
    * Set widget height.
    * 
    * @param   height
    *          New widget height.
    */
   @Override
   public void setHeight(double height)
   {
      forcedSize = true;
      super.setHeight(height);
   }
   
   /**
    * This component will not receive an input focus when Tab or Shift-Tab is
    * pressed.
    * 
    * @return  Always <code>false</code>.
    */
   @Override
   public boolean focusTraversable()
   {
      return false;
   }

   /**
    * Checks whether this attribute container can receive focus generally.
    * 
    * @return  <code>true</code> if this container can be focused under
    *          certain conditions. <code>false</code> if, and only if, this
    *          attribute container can never receive focus.    
    */
   @Override
   public boolean isFocusable()
   {
      return false;
   }

   /**
    * Synchronizes the actual size with the default size.
    */
   protected void updateSize()
   {
      if (forcedSize)
         return;

      String[] values = name().split("\n");

      int width = 0;
      int height = 0;

      for (int i = 0; i < values.length; i++)
      {
         if (width < values[i].length())
            width = values[i].length();
         height++;
      }

      if (underlined)
         height++;

      //Call super to avoid setting forcedSize to true
      super.setSize(width, height);
   }

   /**
    * Set underlining mode.
    * 
    * @param   b
    *          New value for the mode.
    */
   public void setUnderlined(boolean b)
   {
      underlined = b;
      updateSize();
   }

   /**
    * Get current value of the underlining flag.
    * 
    * @return  <code>true</code> if label is underlined.
    */
   public boolean isUnderlined()
   {
      return underlined;
   }
   
   /**
    * Get current value of the alignment.
    * 
    * @return  <code>true</code> if right alignment is active.
    */
   public boolean isRightAligned()
   {
      return rightAligned;
   }
   
   /**
    * Activate or deactivate right alignment of the text.
    * 
    * @param   rightAligned
    *          <code>true</code> activate right alignment,
    *          <code>false</code>- deactivate.
    */
   public void setRightAligned(boolean rightAligned)
   {
      this.rightAligned = rightAligned;
   }
   
   /**
    * Sets the name of the component.
    * 
    * @param   name
    *          New name of the component.
    */
   @Override
   public void setName(String name)
   {
      config.name = name;

      //Reset the forcedSize flag so updateSize() can properly adjust the size.
      forcedSize = false;
      super.setName(name);
      updateSize();
   }

   /**
    * Set extra offset for label.
    * 
    * @param   offset
    *          Size of the shift.
    */
   public void setOffset(double offset)
   {
      this.offset = Coordinate.scale(offset);
   }

   /**
    * Disable empty label mode. Note that actual disabling happens only during
    * next {@link #draw()} call.
    */
   public void setEmptyModeOff()
   {
      disableEmptyMode = true;
   }

   /**
    * Enable empty mode.
    */
   public void setEmptyModeOn()
   {
      emptyMode = true;
   }
   
   /**
    * Process this label so that it's linked with the server-side widget represented by the 
    * <code>serverSideId</code>.
    * 
    * @param    oldLbl
    *           The old label, being replaced by this one.
    * @param    frameId
    *           The ID of the frame to which this label belongs.
    * @param    serverSideId
    *           The server-side ID of this label.
    */
   @SuppressWarnings({ "unchecked", "rawtypes" })
   public void linkTo(Label<O> oldLbl, int frameId, int serverSideId)
   {
      if (serverSideId == -1)
      {
         return;
      }
      
      Point loc = location();

      WidgetConfig lblCfg = config();
      WidgetId oldId = lblCfg.id;
      WidgetId newId = oldId.createId(serverSideId);

      double fixedCol = BaseConfig.INV_COORD;
      double fixedRow = BaseConfig.INV_COORD;
      
      ConfigManager cm = ConfigManager.getInstance();
      WidgetConfig wc = cm.getActiveConfig(newId);
      if (wc instanceof BaseConfig)
      {
         BaseConfig bc = (BaseConfig) wc;
         fixedCol = bc.column;
         fixedRow = bc.row;
      }
      
      double x = loc.x;
      double y = loc.y;
      boolean setLoc = false;

      if (fixedCol != BaseConfig.INV_COORD)
      {
         x = fixedCol;
         setLoc = true;
      }
      if (fixedRow != BaseConfig.INV_COORD)
      {
         y = fixedRow;
         setLoc = true;
      }

      if (setLoc)
      {
         setLocation(x, y);
      }
      
      ClientConfigManager mgr = ConfigManager.getInstance();
      
      // switch the internal ID for this label (from virtual to real)
      mgr.switchId(oldId, newId);
      
      // mark as dirty all configs which are not the same as what the server has...
      mgr.trackAllChanges(lblCfg);
      
      OutputManager.instance().getRegistry().addWidget((Widget) this);
   }
}

