/*
** Module   : LabeledBorderedPanel.java
** Abstract : Panel with border and (optionally) - side or top label.
**
** Copyright (c) 2010-2015, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 SIY 20108020 Refactored common code from RadioSet and SelectionList.
** 002 SIY 20111101 Moved from CHUI package. Minor cleanups/refactorings in preparation to GUI 
**                  client.
** 003 SIY 20111202 Updated imports.
** 004 EVL 20140528 Adding implementation stubs for two new methods.
** 003 CA  20140805 Added generics to force the abstract widget classes to not use implementation-
**                  specific code: the abstract widgets will have access only to the base APIs
**                  provided by the OutputManager.
** 004 HC  20140911 GUI coordinate support - data types refactoring.
** 005 CA  20140926 Added shared widget configuration support, to allow GUI/ChUI concrete
**                  implementation for the same widget ID.  Refactored the widget configuration
**                  classes: all fields were made public; these classes need to be as dumb as
**                  possible, all logic related to setting/getting a certain field should be at
**                  the widget or at the caller. Refs #2254
** 006 CA  20150105 Changes for SIDE-LABEL-HANDLE runtime support.
** 007 CA  20150114 Fixed support for changing the ROW/COLUMN attrs from server-side, for a 
**                  side-label.
** 008 HC  20150128 Added support for frame widget position changes induced by setting
**                  ROW/COL/X/Y attributes.
*/

package com.goldencode.p2j.ui.client;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.chui.ThinClient;
import com.goldencode.p2j.ui.client.border.*;
import com.goldencode.p2j.ui.client.layout.*;
import com.goldencode.p2j.ui.client.widget.*;

/**
 * Common implementation of {@link LabeledWidget} interface for 
 * {@link RadioSet} and {@link SelectionList}.
 * 
 * @author 	SIY
 * @version 1.0
 */
public class LabeledBorderedPanel<O extends OutputManager<?>>
extends BorderedPanel<O>
implements LabeledWidget<O>
{
   /** Component label. */
   private Label<O> label = null;
   
   /**
    * Default constructor.
    */
   public LabeledBorderedPanel()
   {
   }
   
   /**
    * Constructor.
    * 
    * @param   mgr
    *          Layout manager.
    * @param   border
    *          Border for this widget.
    */
   public LabeledBorderedPanel(LayoutManager<O> mgr, Border<O> border)
   {
      super(mgr, border);
   }

   /**
    * Provide access to the attributes stored in the widget.
    * 
    * @return  Reference to attribute container.
    */
   @SuppressWarnings("unchecked")
   @Override
   public ControlSetConfig config()
   {
      return null;
   }
   
   /**
    * Return current value of COLON option.
    * 
    * @return  Current value of the option column.
    */
   @Override
   public double getColon()
   {
      return config().getColon();
   }
   
   /**
    * Return text of the column label.
    * 
    * @return  text of the column label.
    */
   @Override
   public String getColumnLabelText()
   {
      return config().clabel;
   }
   
   /**
    * Return reference to the instance of the <code>Label</code> attached to
    * the widget.
    * 
    * @return  Reference to <code>Label</code> instance.
    */
   @Override
   public Label<O> getLabelInstance()
   {
      return label;
   }

   /**
    * Return text of the label.
    * 
    * @return  text of the label.
    */
   @Override
   public String getLabelText()
   {
      return config().label;
   }

   /**
    * Return TO column.
    * 
    * @return  Column assigned via TO option.
    */
   @Override
   public double getTo()
   {
      return config().getTo();
   }

   /**
    * Get current state of flag which is forces presence of the label
    * regardless from the frame NO-LABELS option.
    * 
    * @return  Current state of the flag.
    */
   @Override
   public boolean isForceLabel()
   {
      return config().forceLabel;
   }
   
   /**
    * Get current state of NO-LABELS flag.
    * 
    * @return  Current state of the flag.
    */
   @Override
   public boolean isNoLabels()
   {
      return !config().labels;
   }
   
   /**
    * Get widget alignment policy.
    * 
    * @return  <code>true</code> if widget is formatted for right alignment.
    */
   @Override
   public boolean isRightAligned()
   {
      return false;
   }

   /**
    * Set new value for the COLON option.
    * 
    * @param   colon
    *          New value for the option column.
    */
   @Override
   public void setColon(double colon)
   {
      config().setColon(colon);
   }

   /**
    * Assign new instance of <code>Label</code> to the widget.
    * 
    * @param   label
    *          Reference to new <code>Label</code> instance.
    */
   @Override
   public void setLabelInstance(Label<O> label)
   {
      int frameId = ThinClient.frameIdFromWidgetId(this.config().id.getId());
      // this call needs to be done before switching the label references.  this is required 
      // because the linkTo call requires knowledge about the previous label associated with 
      // this widget.  note the "label" instance field may have a null reference, in case this is
      // the first assigned label.
      label.linkTo(this.label, frameId, config().sideLabelId);

      this.label = label;
      this.label.setVisible(isVisible());
   }

   /**
    * Force NO-LABELS flag.
    * 
    * @param   labels
    *          New value of the flag.
    */
   @Override
   public void setNoLabels(boolean labels)
   {
      config().labels = !labels;
   }

   /**
    * Assign new TO column.
    * 
    * @param   col
    *          New value for the option.
    */
   @Override
   public void setTo(double col)
   {
      config().setTo(col);
   }


   /**
    * Shows or hides the radio-set and its label.
    *
    * @param    visible
    *           <code>true</code> to make the widget visible, 
    *           <code>false</code> to hide the widget.
    */
   @Override
   public void setVisible(boolean visible)
   {
      super.setVisible(visible);

      if (label != null)
         label.setVisible(visible);
   }
   
   /**
    * Checking if the widget was made visible at least once.
    * 
    * @return  <code>true</code> if the widget was visible,
    *          <code>false</code> otherwise.
    */
   @Override
   public boolean wasVisible()
   {
      return true;
   }
   
   /**
    * Remember the widget was made visible at least once.
    */
   @Override
   public void setWasVisible()
   {
      return;
   }
}
