/*
** Module   : LabeledDataContainer.java
** Abstract : P2JComponent with optional side or top label.
**
** Copyright (c) 2010-2015, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 SIY 20100820 Refactored common code from Editor and FillIn.
** 002 SIY 20100915 Renamed class and base class from LabeledP2JComponent.
** 003 SIY 20111101 Moved from CHUI package. Minor cleanups/refactorings in preparation to GUI 
**                  client.
** 004 SIY 20111202 Updated imports.
** 005 EVL 20140528 Fix for displaying the labels for hidden widgets in a frame.  We need to keep 
**                  track when the widget become visible to differentiate the hidden widgets 
**                  rendering.
** 006 CA  20140805 Added generics to force the abstract widget classes to not use implementation-
**                  specific code: the abstract widgets will have access only to the base APIs
**                  provided by the OutputManager.
** 007 HC  20140911 GUI coordinate support - data types refactoring.
** 008 CA  20140926 Added shared widget configuration support, to allow GUI/ChUI concrete
**                  implementation for the same widget ID.  Refactored the widget configuration
**                  classes: all fields were made public; these classes need to be as dumb as
**                  possible, all logic related to setting/getting a certain field should be at
**                  the widget or at the caller. Refs #2254
** 009 CA  20150105 Changes for SIDE-LABEL-HANDLE runtime support.
** 010 CA  20150114 Fixed support for changing the ROW/COLUMN attrs from server-side, for a 
**                  side-label.
** 011 HC  20150128 Added support for frame widget position changes induced by setting
**                  ROW/COL/X/Y attributes.
*/

package com.goldencode.p2j.ui.client;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.chui.ThinClient;

/**
 * Common implementation of {@link LabeledWidget} interface for 
 * {@link FillIn} and {@link Editor}.
 * 
 * @author  SIY
 * @version 1.0
 */
public abstract class LabeledDataContainer<O extends OutputManager<?>>
extends FixedSizeContainer<O>
implements LabeledWidget<O>,
           DataContainer
{
   /** Component label. */
   private Label<O> sideLabel = null;
   
   /** Flag indicating the widget was made visible at least once. */
   private boolean wasVisible = false;
   
   /**
    * Provide access to the attributes stored in the widget.
    * 
    * @return  Reference to attribute container.
    */
   @SuppressWarnings("unchecked")
   @Override
   public ControlTextConfig config()
   {
      return null;
   }
   
   /**
    * Return current value of COLON attribute.
    * 
    * @return  Current value of the attribute.
    */
   @Override
   public double getColon()
   {
      return config().getColon();
   }

   /**
    * Return text of the column label.
    * 
    * @return  text of the column label.
    */
   @Override
   public String getColumnLabelText()
   {
      return config().clabel;
   }

   /**
    * Return reference to the instance of the <code>Label</code> attached to
    * the widget.
    * 
    * @return  Reference to <code>Label</code> instance.
    */
   @Override
   public Label<O> getLabelInstance()
   {
      return sideLabel;
   }

   /**
    * Return text of the label.
    * 
    * @return  text of the label.
    */
   @Override
   public String getLabelText()
   {
      return config().label;
   }

   /**
    * Return current value of TO attribute.
    * 
    * @return  Current value of the attribute.
    */
   @Override
   public double getTo()
   {
      return config().getTo();
   }

   /**
    * Get current state of flag which is forces presence of the label
    * regardless from the frame NO-LABELS option.
    * 
    * @return  Current state of the flag.
    */
   @Override
   public boolean isForceLabel()
   {
      return config().forceLabel;
   }
   
   /**
    * Get current state of NO-LABELS flag.
    * 
    * @return  Current state of the flag.
    */
   @Override
   public boolean isNoLabels()
   {
      return !config().labels;
   }

   /**
    * Get widget alignment policy.
    * 
    * @return  <code>true</code> if widget is formatted for right alignment.
    */
   @Override
   public boolean isRightAligned()
   {
      return false;
   }

   /**
    * Set new value for the COLON attribute.
    * 
    * @param   colon
    *          New value for the attribute.
    */
   @Override
   public void setColon(double colon)
   {
      config().setColon(colon);
   }

   /**
    * Assign new instance of <code>Label</code> to the fill-in widget.
    * 
    * @param   label
    *          Reference to new <code>Label</code> instance.
    */
   @Override
   public void setLabelInstance(Label<O> label)
   {
      int frameId = ThinClient.frameIdFromWidgetId(this.config().id.getId());
      // this call needs to be done before switching the label references.  this is required 
      // because the linkTo call requires knowledge about the previous label associated with 
      // this widget.  note the "sideLabel" instance field may have a null reference, in case this 
      // is the first assigned label.
      label.linkTo(this.sideLabel, frameId, config().sideLabelId);

      sideLabel = label;
      if (sideLabel.isVisible() != isVisible())
         sideLabel.setVisible(isVisible());
   }
   
   /**
    * Force NO-LABELS flag.
    *
    * @param   labels
    *          New value of the flag.
    */
   @Override
   public void setNoLabels(boolean labels)
   {
      config().labels = !labels;
   }

   /**
    * Set new value for the TO attribute.
    * 
    * @param   pos
    *          New value for the attribute.
    */
   @Override
   public void setTo(double pos)
   {
      config().setTo(pos);
   }
   
   /**
    * Shows or hides the widget and attached label.
    *
    * @param    visible
    *           <code>true</code> to make the widget visible, 
    *           <code>false</code> to hide the widget.
    */
   @Override
   public void setVisible(boolean visible)
   {
      super.setVisible(visible);

      if (sideLabel != null)
         sideLabel.setVisible(visible);
   }

   /**
    * Getting the modification state of the widget.
    * 
    * @return  <code>true</code> if the widget value was modified,
    *          <code>false</code> otherwise.
    */
   @Override
   public boolean isModified()
   {
      return config().modified;
   }
   
   /**
    * Checking if the widget was made visible at least once.
    * 
    * @return  <code>true</code> if the widget was visible,
    *          <code>false</code> otherwise.
    */
   @Override
   public boolean wasVisible()
   {
      return wasVisible;
   }
   
   /**
    * Remember the widget was made visible at least once.
    */
   @Override
   public void setWasVisible()
   {
      wasVisible = true;
   }
}
